# -*- coding: utf-8 -*-
# Copyright (c) 2012-2014, Philip Xu <pyx@xrefactor.com>
# License: BSD New, see LICENSE for details.
"""smonad - a functional library"""

__version__ = (0, 6)

VERSION = '%d.%d' % __version__

__all__ = [
]
